package org.processmining.causalactivitygraphcreator.creators.impl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.causalactivitygraph.models.CausalActivityGraph;
import org.processmining.causalactivitygraph.models.impl.CausalActivityGraphFactory;
import org.processmining.causalactivitygraphcreator.creators.GraphCreator;
import org.processmining.causalactivitygraphcreator.creators.GraphCreatorParameters;
import org.processmining.causalactivitymatrix.models.CausalActivityMatrix;
import org.processmining.framework.plugin.PluginContext;

public class GraphSplitIOCreator implements GraphCreator {

	public final static String NAME = "Binary Split";

	private class IO {
		IO() {
			inputs = new HashSet<XEventClass>();
			outputs = new HashSet<XEventClass>();
		}

		Set<XEventClass> inputs;
		Set<XEventClass> outputs;
		
		public String toString() {
			return "[inputs = " + inputs + ",outputs = " + outputs + "]";
		}
		
		public boolean equals(Object object) {
			if (object instanceof IO) {
				IO io = (IO) object;
				return io.inputs.equals(inputs) && io.outputs.equals(outputs);
			} 
			return false;
		}
	}

	private class Problem {
		Problem() {
			io = new IO();
			cluster1 = new IO();
			cluster2 = new IO();
		}

		IO io;
		IO cluster1;
		IO cluster2;
	}

	private class Solution {
		Solution() {
			score = Double.NEGATIVE_INFINITY;
			cluster1 = new IO();
			cluster2 = new IO();
		}

		double score;
		IO cluster1;
		IO cluster2;
	}

	private Solution bestSolution;

	public CausalActivityGraph apply(PluginContext context, CausalActivityMatrix matrix,
			GraphCreatorParameters parameters) {
		IO initialIO = initialIO(matrix);
		List<XEventClass> activities = matrix.getActivities();
		Set<IO> bestIOs = split(activities, initialIO, matrix, parameters);
		parameters.displayMessage("[GraphSplitIOCreator] Solution = " + bestIOs);
		CausalActivityGraph graph = CausalActivityGraphFactory.createCausalActivityGraph();
		graph.init(matrix.getLabel(), new HashSet<XEventClass>(activities));
		for (XEventClass input : activities) {
			for (XEventClass output : activities) {
				check(input, output, bestIOs, graph, matrix);
			}
		}
		return graph;
	}

	public String getName() {
		return NAME;
	}

	private IO initialIO(CausalActivityMatrix matrix) {
		IO initialIO = new IO();
		List<XEventClass> activities = matrix.getActivities();
		for (XEventClass input : activities) {
			double maxValue = Double.NEGATIVE_INFINITY;
			for (XEventClass output : activities) {
				double value = matrix.getValue(input, output);
				if (value > maxValue) {
					maxValue = value;
				}
			}
			if (maxValue <= -1.0) {
				initialIO.outputs.add(input);
			}
		}
		for (XEventClass output : activities) {
			double maxValue = Double.NEGATIVE_INFINITY;
			for (XEventClass input : activities) {
				double value = matrix.getValue(input, output);
				if (value > maxValue) {
					maxValue = value;
				}
			}
			if (maxValue <= -1.0) {
				initialIO.inputs.add(output);
			}
		}
		return initialIO;
	}

	private void check(XEventClass input, XEventClass output, Set<IO> bestIOs, CausalActivityGraph graph,
			CausalActivityMatrix matrix) {
		double value = (matrix.getValue(input, output));
//		if (value > 0) {
			for (IO bestIO : bestIOs) {
				if (bestIO.inputs.contains(input) && bestIO.outputs.contains(output)) {
					graph.setCausality(input, output, (2.0 + value)/3.0);
					return;
				}
			}
//		}
	}

	private Set<IO> split(List<XEventClass> activities, IO io, CausalActivityMatrix matrix, GraphCreatorParameters parameters) {
		List<IO> workIOs = new ArrayList<IO>();
		Set<IO> bestIOs = new HashSet<IO>();
		Problem problem = new Problem();
		problem.io = io;
		bestSolution = new Solution();
		parameters.displayMessage("[GraphSplitIOCreator] Initial split");
		split(activities, activities.size(), matrix, problem, parameters);
		parameters.displayMessage("[GraphSplitIOCreator] Cluster 1 = " + bestSolution.cluster1);
		parameters.displayMessage("[GraphSplitIOCreator] Cluster 2 = " + bestSolution.cluster2);
		workIOs.add(bestSolution.cluster1);
		workIOs.add(bestSolution.cluster2);
		while (!workIOs.isEmpty()) {
			parameters.displayMessage("[GraphSplitIOCreator] Cluster to do = " + workIOs.size());
			IO workIO = workIOs.remove(0);
			Set<XEventClass> subActivities = new HashSet<XEventClass>();
			subActivities.addAll(workIO.inputs);
			subActivities.addAll(workIO.outputs);
			if (subActivities.size() < 1) {
				parameters.displayMessage("[GraphSplitIOCreator] Remove " + workIO);
				bestIOs.add(workIO);
			} else {
				double score = score(matrix, workIO);
				Problem subProblem = new Problem();
				subProblem.io = new IO();
				subProblem.io.inputs.addAll(workIO.inputs);
				subProblem.io.outputs.addAll(workIO.outputs);
				subProblem.io.inputs.removeAll(workIO.outputs);
				subProblem.io.outputs.removeAll(workIO.inputs);
				bestSolution = new Solution();
				parameters.displayMessage("[GraphSplitIOCreator] Split Cluster " + workIO + ", score = " + score);
				split(new ArrayList<XEventClass>(subActivities), subActivities.size(), matrix, subProblem, parameters);
				double score1 = score(matrix, bestSolution.cluster1);
				double score2 = score(matrix, bestSolution.cluster2);
				parameters.displayMessage("[GraphSplitIOCreator] Cluster 1 = " + bestSolution.cluster1 + ", score = " + score1);
				parameters.displayMessage("[GraphSplitIOCreator] Cluster 2 = " + bestSolution.cluster2 + ", score = " + score2);
				if (accumulate(score1, score2) + 0.000001 >= score) {
					parameters.displayMessage("[GraphSplitIOCreator] Split");
					parameters.displayMessage("[GraphSplitIOCreator] Remove " + workIO);
					workIOs.add(bestSolution.cluster1);
					workIOs.add(bestSolution.cluster2);
				} else {
					parameters.displayMessage("[GraphSplitIOCreator] No split");
					bestIOs.add(workIO);
				}
			}
		}
		return bestIOs;
	}

	private void split(List<XEventClass> activities, int index, CausalActivityMatrix matrix, Problem problem, GraphCreatorParameters parameters) {
//		parameters.displayMessage("[GraphSplitIOCreator] index = " + index);
		if (index == 0) {
			if (valid(problem)) {
				double score = accumulate(score(matrix, problem.cluster1), score(matrix, problem.cluster2));
				if (score > bestSolution.score) {
					parameters.displayMessage("[GraphSplitIOCreator] Best Clusters = [" + problem.cluster1 + "," + problem.cluster2 + "], score = " + score);
					bestSolution.score = score;
					bestSolution.cluster1.inputs.clear();
					bestSolution.cluster1.inputs.addAll(problem.cluster1.inputs);
					bestSolution.cluster1.outputs.clear();
					bestSolution.cluster1.outputs.addAll(problem.cluster1.outputs);
					bestSolution.cluster2.inputs.clear();
					bestSolution.cluster2.inputs.addAll(problem.cluster2.inputs);
					bestSolution.cluster2.outputs.clear();
					bestSolution.cluster2.outputs.addAll(problem.cluster2.outputs);
				}
			}
		} else {
			XEventClass activity = activities.get(--index);
			if (problem.io.inputs.contains(activity)) {
				problem.cluster1.inputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster2.inputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster1.inputs.remove(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster2.inputs.remove(activity);
			} else if (problem.io.outputs.contains(activity)) {
				problem.cluster1.outputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster2.outputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster1.outputs.remove(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster2.outputs.remove(activity);
			} else {
				problem.cluster1.inputs.add(activity);
				problem.cluster1.outputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster1.outputs.remove(activity);
				problem.cluster2.outputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster1.inputs.remove(activity);
				problem.cluster2.inputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster2.outputs.remove(activity);
				problem.cluster1.outputs.add(activity);
				split(activities, index, matrix, problem, parameters);
				problem.cluster2.inputs.remove(activity);
				problem.cluster1.outputs.remove(activity);
			}
		}
	}

	private double accumulate(double score1, double score2) {
		return score1 + score2;
	}

	private double score(CausalActivityMatrix matrix, IO io) {
		if (io.inputs.isEmpty() || io.outputs.isEmpty()) {
			return Double.NEGATIVE_INFINITY;
		}
		double score = 0.0;
		for (XEventClass input : io.inputs) {
			for (XEventClass output : io.outputs) {
				double value = matrix.getValue(input, output);
//				if (value > 0) {
					score += (1.0 + value)/2.0;
//				}
			}
		}
		return score;
	}

	private boolean valid(Problem problem) {
		if (problem.cluster1.inputs.containsAll(problem.cluster2.inputs)) {
			return false;
		}
		if (problem.cluster2.inputs.containsAll(problem.cluster1.inputs)) {
			return false;
		}
		if (problem.cluster1.outputs.containsAll(problem.cluster2.outputs)) {
			return false;
		}
		if (problem.cluster2.outputs.containsAll(problem.cluster1.outputs)) {
			return false;
		}
		return true;
	}
}
